/**
 *    DefaultFrameWriter.java
 *    Copyright (C) 2010 New Zealand Digital Library, http://expeditee.org
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.expeditee.io;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashMap;

import org.expeditee.agents.WriteTree;
import org.expeditee.gio.EcosystemManager;
import org.expeditee.gui.Frame;
import org.expeditee.gui.FrameIO;
import org.expeditee.items.Item;
import org.expeditee.items.Line;
import org.expeditee.items.Text;

public abstract class DefaultFrameWriter extends ItemWriter implements
		FrameWriter {

	public static final String TAB_INDEX = "_ti";
	public static final String SINGLE_LINE_ONLY = "_sl";
	public static final String PLACEHOLDER = "_ph";
	public static final String ENCRYPTION_LABEL = "_el";
	public static final char MAGNETIZED_ITEM_BOTTOM = '/';
	public static final char MAGNETIZED_ITEM_TOP = '^';
	public static final char MAGNETIZED_ITEM_RIGHT = ']';
	public static final char MAGNETIZED_ITEM_LEFT = '[';
	public static final char PERMISSION = 'p';
	public static final char LINK_HISTORY = 'B';
	public static final char TOOLTIP = 'A';
	public static final char CONSTRAINT_IDS = 'c';
	public static final char LINE_IDS = 'l';
	public static final char THICKNESS = 'h';
	public static final char AUTO_WRAP_TO_SAVE = 'r';
	public static final char JUSTIFICATION = 'k';
	public static final char MIN_WIDTH_TO_SAVE = 'M';
	public static final char WIDTH_TO_SAVE = 'w';
	public static final char INITIAL_SPACING = 'm';
	public static final char LETTER_SPACING = 'b';
	public static final char WORD_SPACING = 'a';
	public static final char FORMULA = 'J';
	public static final char MASK = 'O';
	public static final char SPACING = 't';
	public static final char FONT = 'f';
	public static final char FILLED = 'z';
	public static final char DOT_TYPE = 'v';
	public static final char ARROW = 'j';
	public static final char LINE_PATTERN = 'g';
	public static final char LINK_TEMPLATE = 'y';
	public static final char LINK_FRAMESET = 'q';
	public static final char LINK_MARK = 'n';
	public static final char OWNER = 'o';
	public static final char FILL_PATTERN = 'i';
	public static final char ANCHOR_BOTTOM = 'I';
	public static final char ANCHOR_TOP = 'N';
	public static final char ANCHOR_RIGHT = 'H';
	public static final char ANCHOR_LEFT = 'R';
	public static final char GRADIENT_ANGLE = 'Q';
	public static final char GRADIENT_COLOR = 'E';
	public static final char FILL_COLOR = 'e';
	public static final char HIGHLIGHT = 'u';
	public static final char DATA = 'D';
	public static final char ACTION_LEAVE_FRAME = 'Y';
	public static final char ACTION_ENTER_FRAME = 'W';
	public static final char ACTION_CURSOR_LEAVE = 'V';
	public static final char ACTION_CURSOR_ENTER = 'U';
	public static final char ACTION_MARK = 'x';
	public static final char ACTION = 'X';
	public static final char LINK = 'F';
	public static final char TEXT = 'T';
	public static final char POSITION = 'P';
	public static final char BORDER_COLOR = 'K';
	public static final char BACKGROUND_COLOR = 'G';
	public static final char COLOR = 'd';
	public static final char DATE_CREATED = 's';
	public static final char TYPE_AND_ID = 'S';

	protected String _filename = null;

	protected String _output = null;

	protected ProxyWriter _writer = null;

	protected String _format = "";

	protected boolean _running = true;

	protected boolean _stop = false;

	// keep track of methods that are put on the same line
	private static LinkedHashMap<Character, Method> _ItemTags = null;
	// IMPORTANT: keys in _ItemTagsExt must start with underscore as their first character 
	private static LinkedHashMap<String, Method> _ItemTagsExt = null;

	protected static LinkedHashMap<Character, Method> _FrameTags = null;

	public DefaultFrameWriter() {
		
		if (_ItemTags != null && _FrameTags != null)
			return;

		_ItemTags = new LinkedHashMap<Character, Method>();
		_ItemTagsExt = new LinkedHashMap<String, Method>();
		_FrameTags = new LinkedHashMap<Character, Method>();

		try {
			_FrameTags.put('V', Frame.class.getMethod("getVersion"));
			_FrameTags.put('p', Frame.class.getMethod("getPermission"));
			_FrameTags.put('U', Frame.class.getMethod("getOwner"));
			_FrameTags.put('D', Frame.class.getMethod("getDateCreated"));
			_FrameTags.put('M', Frame.class.getMethod("getLastModifyUser"));
			_FrameTags.put('d', Frame.class.getMethod("getLastModifyDate"));
			_FrameTags.put('F', Frame.class.getMethod("getFrozenDate"));
			_FrameTags.put('O', Frame.class.getMethod("getForegroundColor"));
			_FrameTags.put('B', Frame.class.getMethod("getBackgroundColor"));
			_FrameTags.put('K', Frame.class.getMethod("getEncryptionLabel"));
			_FrameTags.put('T', Frame.class.getMethod("getData"));
			_FrameTags.put('E', Frame.class.getMethod("getEncryptionPermission"));
			_FrameTags.put('G', Frame.class.getMethod("getGroup"));
			
			// Note: As of 26/11/18 there are no unused letter item tags.  Use other characters.
			_ItemTags.put(DefaultFrameWriter.TYPE_AND_ID, Item.class.getMethod("getTypeAndID"));
			_ItemTags.put(DefaultFrameWriter.DATE_CREATED, Item.class.getMethod("getDateCreated"));
			_ItemTags.put(DefaultFrameWriter.COLOR, Item.class.getMethod("getColor"));
			_ItemTags.put(DefaultFrameWriter.BACKGROUND_COLOR, Item.class.getMethod("getBackgroundColor"));
			_ItemTags.put(DefaultFrameWriter.BORDER_COLOR, Item.class.getMethod("getBorderColor"));

			_ItemTags.put(DefaultFrameWriter.POSITION, Item.class.getMethod("getPosition"));
			_ItemTags.put(DefaultFrameWriter.TEXT, Text.class.getMethod("getText"));
			_ItemTags.put(DefaultFrameWriter.LINK, Item.class.getMethod("getLink"));
			_ItemTags.put(DefaultFrameWriter.ACTION, Item.class.getMethod("getAction"));
			_ItemTags.put(DefaultFrameWriter.ACTION_MARK, Item.class.getMethod("getActionMark"));
			_ItemTags.put(DefaultFrameWriter.ACTION_CURSOR_ENTER, Item.class.getMethod("getActionCursorEnter"));
			_ItemTags.put(DefaultFrameWriter.ACTION_CURSOR_LEAVE, Item.class.getMethod("getActionCursorLeave"));
			_ItemTags.put(DefaultFrameWriter.ACTION_ENTER_FRAME, Item.class.getMethod("getActionEnterFrame"));
			_ItemTags.put(DefaultFrameWriter.ACTION_LEAVE_FRAME, Item.class.getMethod("getActionLeaveFrame"));
			_ItemTags.put(DefaultFrameWriter.DATA, Item.class.getMethod("getData"));
			_ItemTags.put(DefaultFrameWriter.HIGHLIGHT, Item.class.getMethod("getHighlight"));
			_ItemTags.put(DefaultFrameWriter.FILL_COLOR, Item.class.getMethod("getFillColor"));
			_ItemTags.put(DefaultFrameWriter.GRADIENT_COLOR, Item.class.getMethod("getGradientColor"));
			_ItemTags.put(DefaultFrameWriter.GRADIENT_ANGLE, Item.class.getMethod("getGradientAngle"));
			
			_ItemTags.put(DefaultFrameWriter.ANCHOR_LEFT, Item.class.getMethod("getAnchorLeft"));
			_ItemTags.put(DefaultFrameWriter.ANCHOR_RIGHT, Item.class.getMethod("getAnchorRight"));
			_ItemTags.put(DefaultFrameWriter.ANCHOR_TOP, Item.class.getMethod("getAnchorTop"));
			_ItemTags.put(DefaultFrameWriter.ANCHOR_BOTTOM, Item.class.getMethod("getAnchorBottom"));

			_ItemTags.put(DefaultFrameWriter.FILL_PATTERN, Item.class.getMethod("getFillPattern"));
			_ItemTags.put(DefaultFrameWriter.OWNER, Item.class.getMethod("getOwner"));
			_ItemTags.put(DefaultFrameWriter.LINK_MARK, Item.class.getMethod("getLinkMark"));
			_ItemTags.put(DefaultFrameWriter.LINK_FRAMESET, Item.class.getMethod("getLinkFrameset"));
			_ItemTags.put(DefaultFrameWriter.LINK_TEMPLATE, Item.class.getMethod("getLinkTemplate"));
			_ItemTags.put(DefaultFrameWriter.LINE_PATTERN, Item.class.getMethod("getLinePattern"));

			_ItemTags.put(DefaultFrameWriter.ARROW, Item.class.getMethod("getArrow"));
			
			_ItemTags.put(DefaultFrameWriter.DOT_TYPE, Item.class.getMethod("getDotType"));
			_ItemTags.put(DefaultFrameWriter.FILLED, Item.class.getMethod("getFilled"));

			_ItemTags.put(DefaultFrameWriter.FONT, Text.class.getMethod("getFont"));
			_ItemTags.put(DefaultFrameWriter.SPACING, Text.class.getMethod("getSpacing"));
			
			_ItemTags.put(DefaultFrameWriter.MASK, Text.class.getMethod("getMask"));

			// TODO set a boolean flag to indicate that the text is a formula
			// Store the formula in the text property NOT the answer
			_ItemTags.put(DefaultFrameWriter.FORMULA, Item.class.getMethod("getFormula"));

			_ItemTags.put(DefaultFrameWriter.WORD_SPACING, Text.class.getMethod("getWordSpacing"));
			_ItemTags.put(DefaultFrameWriter.LETTER_SPACING, Text.class.getMethod("getLetterSpacing"));
			_ItemTags.put(DefaultFrameWriter.INITIAL_SPACING, Text.class.getMethod("getInitialSpacing"));
			_ItemTags.put(DefaultFrameWriter.WIDTH_TO_SAVE, Text.class.getMethod("getWidthToSave"));
			_ItemTags.put(DefaultFrameWriter.MIN_WIDTH_TO_SAVE, Text.class.getMethod("getMinWidthToSave"));
			_ItemTags.put(DefaultFrameWriter.JUSTIFICATION, Text.class.getMethod("getJustification"));
			_ItemTags.put(DefaultFrameWriter.AUTO_WRAP_TO_SAVE, Text.class.getMethod("getAutoWrapToSave"));

			_ItemTags.put(DefaultFrameWriter.THICKNESS, Item.class.getMethod("getThickness"));
			_ItemTags.put(DefaultFrameWriter.LINE_IDS, Item.class.getMethod("getLineIDs"));
			_ItemTags.put(DefaultFrameWriter.CONSTRAINT_IDS, Item.class.getMethod("getConstraintIDs"));
			
			_ItemTags.put(DefaultFrameWriter.TOOLTIP, Item.class.getMethod("getTooltip"));
			_ItemTags.put(DefaultFrameWriter.LINK_HISTORY, Item.class.getMethod("getLinkHistory"));
			
			_ItemTags.put(DefaultFrameWriter.PERMISSION, Item.class.getMethod("getPermission"));

			_ItemTags.put(DefaultFrameWriter.MAGNETIZED_ITEM_LEFT, Item.class.getMethod("getMagnetizedItemLeft"));
			_ItemTags.put(DefaultFrameWriter.MAGNETIZED_ITEM_RIGHT, Item.class.getMethod("getMagnetizedItemRight"));
			_ItemTags.put(DefaultFrameWriter.MAGNETIZED_ITEM_TOP, Item.class.getMethod("getMagnetizedItemTop"));
			_ItemTags.put(DefaultFrameWriter.MAGNETIZED_ITEM_BOTTOM, Item.class.getMethod("getMagnetizedItemBottom"));
			
			_ItemTagsExt.put(DefaultFrameWriter.ENCRYPTION_LABEL, Item.class.getMethod("getEncryptionLabel"));
			
			_ItemTagsExt.put(DefaultFrameWriter.PLACEHOLDER, Text.class.getMethod("getPlaceholder"));
			_ItemTagsExt.put(DefaultFrameWriter.SINGLE_LINE_ONLY, Text.class.getMethod("isSingleLineOnly"));
			_ItemTagsExt.put(DefaultFrameWriter.TAB_INDEX, Text.class.getMethod("getTabIndex"));
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public void setOutputLocation(String filename) {
		_filename = filename;
	}

	public String writeFrame(Frame toWrite) throws IOException {
		return writeFrame(toWrite, null);
	}

	public String writeFrame(Frame toWrite, Writer writer) throws IOException {
		try {
			initialise(toWrite, writer);

			outputFrame(toWrite);

			_running = false;
			return finaliseFrame();
		} catch (IOException ioe) {
			_running = false;
			throw ioe;
		}

	}

	/**
	 * Called before writing out the body items of each frame.
	 * 
	 * @param starting
	 *            the name of the frame currently being written out.
	 * @throws IOException
	 */
	protected void writeStartFrame(Frame starting) throws IOException {
		if (starting.getTitleItem() != null) {
			if (starting.getTitleItem().isAnnotation())
				this.writeAnnotationTitle(starting.getTitleItem());
			else
				this.writeTitle(starting.getTitleItem(), starting.getItems());
		}
	}

	/**
	 * Called after writing out the body items of each frame.
	 * 
	 * @param ending
	 *            the name of the frame currently being written out.
	 * @throws IOException
	 */
	protected void writeEndFrame(Frame ending) throws IOException {
	}

	protected final void initialise(Frame start) throws IOException {
		initialise(start, null);
	}

	protected void initialise(Frame start, Writer writer) throws IOException {
		if (_filename == null)
			_filename = FrameIO.EXPORTS_PATH + getFileName(start) + _format;

		if (writer != null) {
			_writer = new ProxyWriter(writer);
			_output = writer.toString();
		} else if (_filename.equalsIgnoreCase(WriteTree.CLIPBOARD)) {
			_writer = new ProxyWriter(true);
			_output = WriteTree.CLIPBOARD;
		} else {
			if (_filename.contains(File.separator)) {
				String extTest = _filename.substring(_filename
						.lastIndexOf(File.separator) + 1);
				if (!extTest.contains("."))
					_filename += _format;
			} else if (!_filename.contains("."))
				_filename += _format;

			if (!_filename.contains(File.separator))
				_filename = FrameIO.EXPORTS_PATH + _filename;

			File test = new File(_filename);

			File parent = test.getParentFile();
			if (parent != null && !parent.exists()) {
				parent.mkdirs();
			}

			// Open an Output Stream Writer to set encoding
			OutputStream fout = new FileOutputStream(_filename);
			OutputStream bout = new BufferedOutputStream(fout);
			Writer out = new OutputStreamWriter(bout, "UTF-8");

			_writer = new ProxyWriter(out);
			_output = _filename;
		}
	}

	protected String getFileName(Frame start) {
		return getValidFilename(start.getTitle());
	}

	public static String getValidFilename(String filename) {
		return filename.replaceAll("[ \\.]", "_");
	}

	protected String finalise() throws IOException {
		try {
			_writer.flush();
			_writer.close();
		} catch (IOException ioe) {
		} finally {
			_writer.close();
		}
		String errorMessage = EcosystemManager.getMiscManager().open(_output);
		if (errorMessage != null) System.err.println(errorMessage);
		return " exported to " + _output;
	}

	protected String finaliseFrame() throws IOException {
		return "Frame" + finalise();
	}

	protected void outputFrame(Frame toWrite) throws IOException {
		writeStartFrame(toWrite);

		Collection<Item> done = new HashSet<Item>();

		for (Item i : getItemsToWrite(toWrite)) {
			if (_stop) {
				return;
			}

			if (i instanceof Line) {
				if (done.contains(i)) {
					continue;
				}
				done.addAll(i.getAllConnected());
			}
			writeItem(i);
		}
		writeEndFrame(toWrite);
	}

	protected Collection<Item> getItemsToWrite(Frame toWrite) {
		return toWrite.getItemsToSave();
	}

	public boolean isRunning() {
		return _running;
	}

	public void stop() {
		_stop = true;
	}

	public String getFileContents() {
		return "Not Supported";
	}

	/**
	 * @return the _ItemTags
	 */
	public static LinkedHashMap<Character, Method> getItemTags() {
		return _ItemTags;
	}

	/**
	 * @return the _ItemTagsExt
	 */
	public static LinkedHashMap<String, Method> getItemTagsExt() {
		return _ItemTagsExt;
	}

}
